# Authors: Isak Samsten
# License: BSD 3 clause
import numpy as np
from sklearn.pipeline import make_pipeline, make_union
from sklearn.utils import check_random_state

from ..datasets.preprocess import SparseScaler
from ..transform import DiffTransform, HydraTransform
from ._transform import TransformRidgeClassifierCV


class HydraClassifier(TransformRidgeClassifierCV):
    """
    A Dictionary based method using convolutional kernels.

    Parameters
    ----------
    n_groups : int, optional
        The number of groups of kernels.
    n_kernels : int, optional
        The number of kernels per group.
    kernel_size : int, optional
        The size of the kernel.
    sampling : {"normal"}, optional
        The strategy for sampling kernels. By default kernel weights
        are sampled from a normal distribution with zero mean and unit
        standard deviation.
    sampling_params : dict, optional
        Parameters to the sampling approach. The "normal" sampler
        accepts two parameters: `mean` and `scale`.
    order : int, optional
        The order of difference. If set, half the groups with corresponding
        kernels will be convolved with the `order` discrete difference along
        the time dimension.
    alphas : array-like of shape (n_alphas,), optional
        Array of alpha values to try.
    fit_intercept : bool, optional
        Whether to calculate the intercept for this model.
    scoring : str, callable, optional
        A string or a scorer callable object with signature
        `scorer(estimator, X, y)`.
    cv : int, cross-validation generator or an iterable, optional
        Determines the cross-validation splitting strategy.
    class_weight : dict or 'balanced', optional
        Weights associated with classes in the form `{class_label: weight}`.
    normalize : bool, optional
        Standardize before fitting. By default use
        :class:`datasets.preprocess.SparseScaler` to standardize the attributes. Set
        to `False` to disable or `True` to use `StandardScaler`.
    n_jobs : int, optional
        The number of jobs to run in parallel. A value of `None` means using
        a single core and a value of `-1` means using all cores. Positive
        integers mean the exact number of cores.
    random_state : int or RandomState, optional
        Controls the random resampling of the original dataset.

        - If `int`, `random_state` is the seed used by the random number
          generator.
        - If :class:`numpy.random.RandomState` instance, `random_state` is
          the random number generator.
        - If `None`, the random number generator is the
          :class:`numpy.random.RandomState` instance used by
          :func:`numpy.random`.

    References
    ----------
    Dempster, A., Schmidt, D. F., & Webb, G. I. (2023).
        Hydra: competing convolutional kernels for fast and accurate
        time series classification. Data Mining and Knowledge Discovery
    """

    def __init__(
        self,
        *,
        n_groups=64,
        n_kernels=8,
        kernel_size=9,
        sampling="normal",
        sampling_params=None,
        order=1,
        alphas=(0.1, 1.0, 10.0),
        fit_intercept=True,
        scoring=None,
        cv=None,
        class_weight=None,
        normalize="sparse",
        n_jobs=None,
        random_state=None,
    ):
        super().__init__(
            alphas=alphas,
            fit_intercept=fit_intercept,
            scoring=scoring,
            cv=cv,
            class_weight=class_weight,
            n_jobs=n_jobs,
            normalize=normalize,
            random_state=random_state,
        )
        self.n_groups = n_groups
        self.n_kernels = n_kernels
        self.kernel_size = kernel_size
        self.sampling = sampling
        self.sampling_params = sampling_params
        self.order = order

    def _build_pipeline(self):
        pipeline = super()._build_pipeline()
        if self.normalize == "sparse":
            pipeline[1] = ("normalize", SparseScaler())

        return pipeline

    def _get_transform(self, random_state):
        if self.order is not None and self.order > 0 and self.n_groups > 1:
            random_state = check_random_state(self.random_state)
            return make_union(
                HydraTransform(
                    n_groups=self.n_groups // 2,
                    n_kernels=self.n_kernels,
                    kernel_size=self.kernel_size,
                    sampling=self.sampling,
                    sampling_params=self.sampling_params,
                    random_state=random_state.randint(np.iinfo(np.int32).max),
                    n_jobs=self.n_jobs,
                ),
                make_pipeline(
                    DiffTransform(order=self.order),
                    HydraTransform(
                        n_groups=self.n_groups // 2,
                        n_kernels=self.n_kernels,
                        kernel_size=self.kernel_size,
                        sampling=self.sampling,
                        sampling_params=self.sampling_params,
                        random_state=random_state.randint(np.iinfo(np.int32).max),
                        n_jobs=self.n_jobs,
                    ),
                ),
            )

        else:
            return HydraTransform(
                n_groups=self.n_groups,
                n_kernels=self.n_kernels,
                kernel_size=self.kernel_size,
                sampling=self.sampling,
                sampling_params=self.sampling_params,
                random_state=random_state,
                n_jobs=self.n_jobs,
            )



from pathlib import Path

from docutils.parsers.rst.directives.images import Image


def light_dark_paths(path: Path):
    file = Path(path.parts[-1])
    path = Path(*path.parts[:-1])
    ext = file.suffix
    name = file.stem

    return path.joinpath(f"{name}-light{ext}"), path.joinpath(f"{name}-dark{ext}")


class LightDarkImage(Image):
    def run(self):
        image_path = self.arguments[0]
        light, dark = light_dark_paths(Path(image_path))
        self.arguments[0] = str(light)
        classes = self.options.get("classes", [])
        self.options["classes"] = classes + ["only-light"]

        ret = super().run()
        self.arguments[0] = str(dark)
        self.options["classes"] = classes + ["only-dark"]
        return ret + super().run()


def setup(app):
    setup.app = app
    app.add_directive("ldimage", LightDarkImage)
    return {
        "version": "0.1",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



# -*- coding: utf-8 -*-

import os

from setuptools import setup
from setuptools.extension import Extension

BUILD_ARGS = {
    "default": {
        "posix": {
            "extra_compile_args": ["-O2"],
            "extra_link_args": [],
            "libraries": [],
        },
        "nt": {
            "extra_compile_args": ["/O2"],
            "extra_link_args": [],
            "libraries": [],
        },
    },
    "debug": {
        "posix": {
            "extra_compile_args": ["-O0", "-g"],
            "extra_link_args": [],
            "libraries": [],
        },
        "nt": {
            "extra_compile_args": ["/Od"],
            "extra_link_args": [],
            "libraries": [],
        },
    },
    "optimize": {
        "posix": {
            "extra_compile_args": [
                "-O3",
                "-march=native",
                "-ffast-math",
            ],
            "extra_link_args": [],
            "libraries": [],
        },
        "nt": {
            "extra_compile_args": [
                "/Ox",
                "/fp:fast",
            ],
            "extra_link_args": [],
            "libraries": [],
        },
    },
}


def make_extension(name, extension, defaults):
    import numpy

    include_dirs = extension.get("include_dirs", [])
    include_dirs.append(numpy.get_include())

    libraries = extension.get("libraries", [])
    libraries.extend(defaults["libraries"])

    extra_compile_args = extension.get("extra_compile_args", [])
    extra_compile_args.extend(defaults["extra_compile_args"])

    extra_link_args = extension.get("extra_link_args", [])
    extra_link_args.extend(defaults["extra_compile_args"])

    return Extension(
        name,
        sources=extension["sources"],
        include_dirs=include_dirs,
        libraries=libraries,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
    )


if __name__ == "__main__":
    from Cython.Build import cythonize

    build_type = os.getenv("WILDBOAR_BUILD", "default")
    build_nthreads = int(os.getenv("WILDBOAR_BUILD_NTHREADS", "1"))
    build_args = BUILD_ARGS.get(build_type, {}).get(os.name, None)

    if build_args is None:
        raise RuntimeError(f"{build_type} is not a valid build type")

    external_extra_compile_args = os.getenv("WILDBOAR_EXTRA_COMPILE_ARGS")
    if external_extra_compile_args is not None:
        import re

        build_args["extra_compile_args"].extend(
            re.split(r"\s+", external_extra_compile_args)
        )

    if os.name == "posix":
        build_args["libraries"].append("m")

    extensions = {
        "wildboar.distance.*": {
            "sources": ["src/wildboar/distance/*.pyx"],
        },
        "wildboar.tree.*": {
            "sources": ["src/wildboar/tree/*.pyx"],
        },
        "wildboar.transform.*": {
            "sources": ["src/wildboar/transform/*.pyx"],
        },
        "wildboar.transform.catch22._catch22": {
            "sources": [
                "src/wildboar/transform/catch22/_catch22.pyx",
                "src/wildboar/transform/catch22/src/catch22.c",
            ],
            "include_dirs": ["src/wildboar/transform/catch22/src/"],
        },
        "wildboar.utils._fft._pocketfft": {
            "sources": [
                "src/wildboar/utils/_fft/_pocketfft.pyx",
                "src/wildboar/utils/_fft/src/pocketfft.c",
            ],
            "include_dirs": ["src/wildboar/utils/_fft/src/"],
        },
        "wildboar.utils.*": {"sources": ["src/wildboar/utils/*.pyx"]},
    }
    ext_modules = [
        make_extension(name, options, build_args)
        for name, options in extensions.items()
    ]

    setup(
        ext_modules=cythonize(
            ext_modules,
            nthreads=build_nthreads,
            annotate=build_type == "debug",
        )
    )



# Authors: Isak Samsten
# License: BSD 3 clause
"""Utilities for preprocessing time series."""

import numbers
from functools import partial

import numpy as np
from sklearn.base import OneToOneFeatureMixin, TransformerMixin
from sklearn.utils._param_validation import Interval

from .. import iseos
from ..base import BaseEstimator
from ..transform._sax import piecewice_aggregate_approximation
from ..utils.validation import check_array, check_option

__all__ = [
    "standardize",
    "minmax_scale",
    "maxabs_scale",
    "truncate",
    "named_preprocess",
]


def named_preprocess(name):
    """
    Get a named preprocessor.

    Parameters
    ----------
    name : str
        The name of the preprocessor.

    Returns
    -------
    callable
        The preprocessor function.
    """
    return check_option(_PREPROCESS, name, "name")


def standardize(x):
    """
    Scale x along the time dimension.

    The resulting array will have zero mean and unit standard deviation.

    Parameters
    ----------
    x : ndarray of shape (n_samples, n_timestep) or (n_samples, n_dims, n_timestep)
        The samples.

    Returns
    -------
    ndarray of shape (n_samples, n_timestep) or (n_samples, n_dims, n_timestep)
        The standardized samples.
    """
    x = check_array(x, allow_3d=True, force_all_finite="allow-nan")
    return (x - np.nanmean(x, axis=-1, keepdims=True)) / np.nanstd(
        x, axis=-1, keepdims=True
    )


def minmax_scale(x, min=0, max=1):
    """
    Scale x along the time dimension.

    Each time series is scaled such that each value is between min and max.

    Parameters
    ----------
    x : ndarray of shape (n_samples, n_timestep) or (n_samples, n_dims, n_timestep)
        The samples.
    min : float, optional
        The minimum value.
    max : float, optional
        The maximum value.

    Returns
    -------
    ndarray of shape (n_samples, n_timestep) or (n_samples, n_dims, n_timestep)
        The transformed samples.
    """
    if min > max:
        raise ValueError("min must be smaller than max.")
    x = check_array(x, allow_3d=True, force_all_finite="allow-nan")
    x_min = np.nanmin(x, axis=-1, keepdims=True)
    x_max = np.nanmax(x, axis=-1, keepdims=True)
    x = (x - x_min) / (x_max - x_min)
    return x * (max - min) + min


def maxabs_scale(x):
    """
    Scale each time series by its maximum absolute value.

    Parameters
    ----------
    x : ndarray of shape (n_samples, n_timestep) or (n_samples, n_dims, n_timestep)
        The samples.

    Returns
    -------
    ndarray of shape (n_samples, n_timestep) or (n_samples, n_dims, n_timestep)
        The transformed samples.
    """
    x = check_array(x, allow_3d=True, force_all_finite="allow-nan")
    x_max = np.nanmax(np.abs(x), axis=-1, keepdims=True)
    return x / x_max


def truncate(x, n_shortest=None):
    """
    Truncate x to the shortest sequence.

    Parameters
    ----------
    x : ndarray of shape (n_samples, n_timestep) or (n_samples, n_dims, n_timestep)
        The samples.
    n_shortest : int, optional
        The maximum size.

    Returns
    -------
    ndarray of shape (n_samples, n_shortest) or (n_samples, n_dims, n_shortest)
        The truncated samples.
    """
    x = check_array(x, allow_3d=True, allow_eos=True, force_all_finite="allow-nan")
    if n_shortest is None:
        eos = np.nonzero(iseos(x))[-1]
        if eos.size > 0:
            return x[..., : np.min(eos)]
        else:
            return x
    else:
        if n_shortest > x.shape[-1]:
            raise ValueError("n_shortest > x.shape[-1]")
        return x[..., :n_shortest]


_PREPROCESS = {
    "standardize": standardize,
    "normalize": standardize,
    "minmax_scale": minmax_scale,
    "maxabs_scale": maxabs_scale,
    "truncate": truncate,
    "downsample": piecewice_aggregate_approximation,
    "downsample-25": partial(piecewice_aggregate_approximation, n_intervals=0.25),
    "downsample-50": partial(piecewice_aggregate_approximation, n_intervals=0.5),
}


class SparseScaler(OneToOneFeatureMixin, TransformerMixin, BaseEstimator):
    """
    Scale attributes.

    The attributes are scaled by:::

        sqrt(x) - mean(sqrt(x), axis=0) / std(sqrt(x), axis=0) + epsilon

    where `epsilon` is given by ``mean(sqrt(x) == 0, axis=0) ** exp``

    Parameters
    ----------
    mask_zero : bool, optional
        Keep zero values at zero.
    exp : float, optional
        The exponent of the feature sparcity.
    """

    _parameter_constraints: dict = {
        "mask_zero": [bool],
        "exp": Interval(numbers.Real, 0, None, closed="left"),
    }

    def __init__(self, mask_zero=True, exp=4):
        self.exp = exp
        self.mask_zero = mask_zero

    def fit(self, x, y=None):
        x = np.sqrt(self._validate_data(x, allow_3d=False).clip(min=0))

        self.mu_ = x.mean(axis=0)
        self.sigma_ = x.std(axis=0) + (x == 0).mean() ** self.exp + 1e-8
        return self

    def transform(self, x):
        x = np.sqrt(self._validate_data(x, allow_3d=False).clip(min=0))

        x = x - self.mu_
        if self.mask_zero:
            x *= x != 0

        x /= self.sigma_

        return x



import numpy as np
from light_dark import mk_light_dark
from matplotlib.cm import get_cmap
from wildboar import iseos
from wildboar.datasets import load_dataset, load_gun_point
from wildboar.datasets.preprocess import minmax_scale, truncate


@mk_light_dark("guide", "datasets", "repository", "preprocess", "minmax_scale.svg")
def gen_preprocess_minmax_scale(fig, ax):
    x, y = load_gun_point()
    ax.plot(x[0])
    x = minmax_scale(x)
    ax.plot(x[0])
    fig.legend(["Original", "Scaled"])


def plot_dim_len(x, fig, ax):
    eos = iseos(x).argmax(axis=2)
    cmap = get_cmap(lut=eos.shape[1])
    for dim in range(eos.shape[1]):
        eos[eos[:, dim] == 0] = x.shape[-1]  # if eos == n_timestep
        ax[dim].scatter(
            np.arange(eos.shape[0]),
            eos[:, dim],
            label="dim %d" % dim,
            marker="x",
            color=cmap(dim),
        )
        ax[dim].set_ylabel(f"dim {dim}")
    fig.tight_layout()


@mk_light_dark(
    "guide", "datasets", "repository", "preprocess", "no-truncate.svg", nrows=3
)
def gen_prerocess_no_truncate(fig, ax):
    x, y = load_dataset(
        "SpokenArabicDigits", repository="wildboar/ucrmts", progress=False
    )
    x = x[0:25, :3, :]
    plot_dim_len(x, fig, ax)


@mk_light_dark("guide", "datasets", "repository", "preprocess", "truncate.svg", nrows=3)
def gen_prerocess_truncate(fig, ax):
    x, y = load_dataset(
        "SpokenArabicDigits", repository="wildboar/ucrmts", progress=False
    )
    x = x[0:25, :3, :]
    plot_dim_len(truncate(x), fig, ax)



"""Evaluation metrics."""

from ._cluster import silhouette_samples, silhouette_score
from ._counterfactual import (
    compactness_score,
    plausability_score,
    proximity_score,
    redudancy_score,
    relative_proximity_score,
    validity_score,
)

__all__ = [
    # Counterfactuals
    "proximity_score",
    "relative_proximity_score",
    "compactness_score",
    "validity_score",
    "plausability_score",
    "redudancy_score",
    # Cluster
    "silhouette_samples",
    "silhouette_score",
]



import numpy as np
import pytest
import wildboar
from wildboar.utils.variable_len import (
    eos,
    get_variable_length,
    is_end_of_series,
    is_variable_length,
)


def test_is_variable_len():
    x1 = np.array([1, 2, 3, eos])
    assert is_variable_length(x1)
    assert is_variable_length(x1.astype(np.float32))

    x2 = np.array([1, 2, 3, eos], dtype=np.float32)
    assert is_variable_length(x2)
    assert is_variable_length(x2.astype(np.float64))


@pytest.mark.parametrize(
    "x",
    [
        np.array(
            [
                [eos, eos, np.nan, np.inf],
                [1, eos, -np.inf, eos],
            ],
            dtype=np.float64,
        ),
    ],
)
def test_is_end_of_series(x):
    expected = np.array(
        [
            [True, True, False, False],
            [False, True, False, True],
        ]
    )
    np.testing.assert_equal(is_end_of_series(x), expected)
    np.testing.assert_equal(is_end_of_series(x.astype(np.float32)), expected)
    # TODO(1.3): Remove
    expected[1, 2] = True
    np.testing.assert_equal(wildboar.iseos(x), expected)


def test_is_end_of_series_2d():
    x = np.array([[1, 2, 3, eos], [1, 2, eos, eos], [1, 2, 3, 4]])
    lengths = get_variable_length(x)
    np.testing.assert_equal(lengths, np.array([3, 2, 4]))


def test_is_end_of_series_1d():
    assert get_variable_length(np.array([1, 2, 3, eos])) == 3
    assert get_variable_length([eos, 0, 0, 0]) == 0


def test_is_end_of_series_3d():
    x = np.array(
        [
            [[1, np.nan, 3, eos], [1, 2, eos, eos], [1, 2, 3, 4]],
            [[1, 2, 3, eos], [1, 2, eos, eos], [eos, eos, 3, 4]],
        ]
    )
    lengths = get_variable_length(x)
    np.testing.assert_equal(lengths, np.array([[3, 2, 4], [3, 2, 0]]))



# Authors: Isak Samsten
# License: BSD 3 clause

import math
import numbers
import warnings

from sklearn.utils._param_validation import Interval, StrOptions

from ._base import BaseAttributeTransform
from ._cinterval import (
    Catch22Summarizer,
    IntervalAttributeGenerator,
    MeanSummarizer,
    MeanVarianceSlopeSummarizer,
    PyFuncSummarizer,
    RandomFixedIntervalAttributeGenerator,
    RandomIntervalAttributeGenerator,
    SlopeSummarizer,
    VarianceSummarizer,
)

_SUMMARIZER = {
    "mean_var_slope": MeanVarianceSlopeSummarizer,
    "mean": MeanSummarizer,
    "variance": VarianceSummarizer,
    "slope": SlopeSummarizer,
    "catch22": Catch22Summarizer,
}


# noqa: H0002
class IntervalMixin:
    """
    Mixin for interval based estimators.

    It provides an implementation for the `_get_generator` method
    which supports interval based transformation.

    The implementing class must have the following properties:

    - `n_intervals`
    - `intervals`
    - `min_size`
    - `max_size`
    - `sample_size`
    - `summarizer`

    See :class:`transform.IntervalTransform` for information about the
    properties.
    """

    _parameter_constraints: dict = {
        "n_intervals": [
            Interval(numbers.Integral, 1, None, closed="left"),
            Interval(numbers.Real, 0, 1, closed="right"),
            StrOptions({"sqrt", "log2"}),
        ],
        "intervals": [
            StrOptions({"fixed", "sample", "random"}),
        ],
        "min_size": [
            Interval(numbers.Real, 0, 1, closed="both"),
            None,
        ],
        "max_size": [
            Interval(numbers.Real, 0, 1, closed="both"),
            None,
        ],
        "sample_size": [Interval(numbers.Real, 0, 1, closed="both")],
        "summarizer": [StrOptions(_SUMMARIZER.keys()), list],
    }

    def _get_generator(self, x, y):  # noqa: PLR0912
        if isinstance(self.summarizer, list):
            if not all(callable(func) for func in self.summarizer):
                raise ValueError(
                    "summarizer must be list of callable or str, got %r"
                    % self.summarizer
                )
            summarizer = PyFuncSummarizer(self.summarizer)
        else:
            summarizer = _SUMMARIZER[self.summarizer]()

        if self.n_intervals == "sqrt":
            n_intervals = math.ceil(math.sqrt(self.n_timesteps_in_))
        elif self.n_intervals in ("log2", "log"):
            # TODO(1.4) Remove
            if self.n_intervals == "log":
                warnings.warn(
                    "The value 'log' for `n_intervals` has been renamed to 'log2' "
                    "and will be removed in 1.4",
                    DeprecationWarning,
                )

            n_intervals = math.ceil(math.log2(self.n_timesteps_in_))
        elif isinstance(self.n_intervals, numbers.Integral):
            n_intervals = self.n_intervals
        else:
            n_intervals = math.ceil(self.n_intervals * self.n_timesteps_in_)

        if self.intervals == "fixed":
            return IntervalAttributeGenerator(n_intervals, summarizer)
        elif self.intervals == "sample":
            return RandomFixedIntervalAttributeGenerator(
                n_intervals, summarizer, math.floor(self.sample_size * n_intervals)
            )
        else:  # "random"
            min_size = self.min_size if self.min_size is not None else 0
            max_size = self.max_size if self.max_size is not None else 1

            if min_size > max_size:
                raise ValueError(
                    f"The min_size parameter of {type(self).__qualname__} "
                    "must be <= max_size."
                )

            min_size = int(self.min_size * self.n_timesteps_in_)
            max_size = int(self.max_size * self.n_timesteps_in_)
            if min_size < 2:
                if self.n_timesteps_in_ < 2:
                    min_size = 1
                else:
                    min_size = 2

            return RandomIntervalAttributeGenerator(
                n_intervals, summarizer, min_size, max_size
            )


class IntervalTransform(IntervalMixin, BaseAttributeTransform):
    """
    Embed a time series as a collection of features per interval.

    Parameters
    ----------
    n_intervals : str, int or float, optional
        The number of intervals to use for the transform.

        - if "log2", the number of intervals is `log2(n_timestep)`.
        - if "sqrt", the number of intervals is `sqrt(n_timestep)`.
        - if int, the number of intervals is `n_intervals`.
        - if float, the number of intervals is `n_intervals * n_timestep`, with
          `0 < n_intervals < 1`.

        .. deprecated:: 1.2
            The option "log" has been renamed to "log2".
    intervals : str, optional
        The method for selecting intervals.

        - if "fixed", `n_intervals` non-overlapping intervals.
        - if "sample", `n_intervals * sample_size` non-overlapping intervals.
        - if "random", `n_intervals` possibly overlapping intervals of randomly
          sampled in `[min_size * n_timestep, max_size * n_timestep]`.
    sample_size : float, optional
        The sample size of fixed intervals if `intervals="sample"`.
    min_size : float, optional
        The minimum interval size if `intervals="random"`.
    max_size : float, optional
        The maximum interval size if `intervals="random"`.
    summarizer : str or list, optional
        The method to summarize each interval.

        - if str, the summarizer is determined by `_SUMMARIZERS.keys()`.
        - if list, the summarizer is a list of functions `f(x) -> float`, where
          `x` is a numpy array.

        The default summarizer summarizes each interval as its mean, standard
        deviation and slope.
    n_jobs : int, optional
        The number of cores to use on multi-core.
    random_state : int or RandomState
        - If `int`, `random_state` is the seed used by the random number generator
        - If `RandomState` instance, `random_state` is the random number generator
        - If `None`, the random number generator is the `RandomState` instance used
          by `np.random`.

    Notes
    -----
    Paralellization dependes on releasing the global interpreter lock (GIL). As
    such, custom functions as summarizers reduces the performance. Wildboar
    implements summarizers for taking the mean ("mean"), variance ("variance")
    and slope ("slope") as well as their combination ("mean_var_slope") and the
    full suite of `catch22` features ("catch22"). In the future, we will allow
    downstream projects to implement their own summarizers in Cython which will
    allow for releasing the GIL.

    References
    ----------
    Lubba, Carl H., Sarab S. Sethi, Philip Knaute, Simon R. Schultz, \
            Ben D. Fulcher, and Nick S. Jones.
        catch22: Canonical time-series characteristics.
        Data Mining and Knowledge Discovery 33, no. 6 (2019): 1821-1852.

    Examples
    --------
    >>> from wildboar.datasets import load_dataset
    >>> x, y = load_dataset("GunPoint")
    >>> t = IntervalTransform(n_intervals=10, summarizer="mean")
    >>> t.fit_transform(x)

    Each interval (15 timepoints) are transformed to their mean.

    >>> t = IntervalTransform(n_intervals="sqrt", summarizer=[np.mean, np.std])
    >>> t.fit_transform(x)

    Each interval (`150 // 12` timepoints) are transformed to two features. The
    mean and the standard deviation.
    """

    _parameter_constraints: dict = {
        **IntervalMixin._parameter_constraints,
        **BaseAttributeTransform._parameter_constraints,
    }

    def __init__(
        self,
        n_intervals="sqrt",
        *,
        intervals="fixed",
        sample_size=0.5,
        min_size=0.0,
        max_size=1.0,
        summarizer="mean_var_slope",
        n_jobs=None,
        random_state=None,
    ):
        super().__init__(n_jobs=n_jobs, random_state=random_state)
        self.n_intervals = n_intervals
        self.summarizer = summarizer
        self.intervals = intervals
        self.sample_size = sample_size
        self.min_size = min_size
        self.max_size = max_size


class FeatureTransform(IntervalTransform):
    """
    Transform a time series as a number of features.

    Parameters
    ----------
    summarizer : str or list, optional
        The method to summarize each interval.

        - if str, the summarizer is determined by `_SUMMARIZERS.keys()`.
        - if list, the summarizer is a list of functions f(x) -> float, where x
          is a numpy array.

        The default summarizer summarizes each time series using `catch22`-features.
    n_jobs : int, optional
        The number of cores to use on multi-core.

    Examples
    --------
    >>> from wildboar.datasets import load_gun_point
    >>> X, y = load_gun_point()
    >>> X_t = FeatureTransform().fit_transform(X)
    >>> X_t[0]
    array([-5.19633603e-01, -6.51047206e-01,  1.90000000e+01,  4.80000000e+01,
            7.48441896e-01, -2.73293560e-05,  2.21476510e-01,  4.70000000e+01,
            4.00000000e-02,  0.00000000e+00,  2.70502518e+00,  2.60000000e+01,
            6.42857143e-01,  1.00000000e-01, -3.26666667e-01,  9.89974643e-01,
            2.90000000e+01,  1.31570726e+00,  1.50000000e-01,  8.50000000e-01,
            4.90873852e-02,  1.47311800e-01])
    """

    _parameter_constraints: dict = {
        "summarizer": IntervalMixin._parameter_constraints["summarizer"],
        "n_jobs": [None, numbers.Integral],
    }

    def __init__(
        self,
        *,
        summarizer="catch22",
        n_jobs=None,
    ):
        super().__init__(n_intervals=1, summarizer=summarizer, n_jobs=n_jobs)



# Authors: Isak Samsten
# License: BSD 3 clause
from abc import ABCMeta, abstractmethod
from numbers import Integral

import numpy as np
from sklearn.base import TransformerMixin
from sklearn.utils.validation import check_is_fitted, check_random_state

from ..base import BaseEstimator
from ..utils.validation import _check_ts_array
from ._attribute_transform import (
    fit,
    fit_transform,
    transform,
)


class BaseAttributeTransform(TransformerMixin, BaseEstimator, metaclass=ABCMeta):
    """
    Base feature engineer transform.

    Parameters
    ----------
    random_state : int or RandomState, optional
        Controls the random resampling of the original dataset.

        - If `int`, `random_state` is the seed used by the
          random number generator.
        - If :class:`numpy.random.RandomState` instance, `random_state` is the
          random number generator.
        - If `None`, the random number generator is the
          :class:`numpy.random.RandomState` instance used by
          :func:`numpy.random`.
    n_jobs : int, optional
        The number of jobs to run in parallel. A value of `None` means using a
        single core and a value of `-1` means using all cores. Positive
        integers mean the exact number of cores.

    Attributes
    ----------
    embedding_ : Embedding
        The underlying embedding.
    """

    _parameter_constraints: dict = {
        "random_state": ["random_state"],
        "n_jobs": [None, Integral],
    }

    def __init__(self, *, random_state=None, n_jobs=None):
        self.n_jobs = n_jobs
        self.random_state = random_state

    def fit(self, x, y=None):
        """
        Fit the transform.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_timestep) or\
                (n_samples, n_dimensions, n_timestep)
            The time series dataset.
        y : None, optional
            For compatibility.

        Returns
        -------
        BaseAttributeTransform
            This object.
        """
        self._validate_params()
        x = self._validate_data(x, allow_3d=True, dtype=np.double)
        self.embedding_ = fit(
            self._get_generator(x, y),
            _check_ts_array(x),
            check_random_state(self.random_state),
        )
        return self

    def transform(self, x):
        """
        Transform the dataset.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_timestep) or\
                (n_samples, n_dimensions, n_timestep)
            The time series dataset.

        Returns
        -------
        ndarray of shape (n_samples, n_outputs)
            The transformation.
        """
        check_is_fitted(self, attributes="embedding_")
        x = self._validate_data(x, reset=False, allow_3d=True, dtype=np.double)
        return transform(self.embedding_, _check_ts_array(x), self.n_jobs)

    def fit_transform(self, x, y=None):
        """
        Fit the embedding and return the transform of x.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_timestep) or\
                (n_samples, n_dimensions, n_timestep)
            The time series dataset.
        y : None, optional
            For compatibility.

        Returns
        -------
        ndarray of shape (n_samples, n_outputs)
            The embedding.
        """
        self._validate_params()
        x = self._validate_data(x, allow_3d=True, dtype=np.double)
        embedding, x_out = fit_transform(
            self._get_generator(x, y),
            _check_ts_array(x),
            check_random_state(self.random_state),
            self.n_jobs,
        )
        self.embedding_ = embedding
        return x_out

    @abstractmethod
    def _get_generator(self, x, y):
        pass



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: language_level=3
# cython: initializedcheck=False

# Authors: Isak Samsten
# License: BSD 3 clause

import numpy as np

from libc.math cimport INFINITY, floor, log2, pow
from libc.stdlib cimport free, malloc
from numpy cimport uint32_t

from ..utils cimport TSArray
from ..utils._misc cimport to_ndarray_int
from ..utils._rand cimport rand_int, rand_normal, rand_uniform
from ._attr_gen cimport Attribute, AttributeGenerator


cdef struct Rocket:
    Py_ssize_t length
    Py_ssize_t dilation
    Py_ssize_t padding
    bint return_mean
    double bias
    double *weight

cdef class KernelSampler:

    cdef void sample(
        self,
        TSArray X,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        double *weights,
        Py_ssize_t length,
        double *mean,
        uint32_t *seed
    ) noexcept nogil:
        pass


cdef class NormalKernelSampler(KernelSampler):
    cdef double mean
    cdef double scale

    def __init__(self, mean=0.0, scale=1.0):
        self.mean = mean
        self.scale = scale

    cdef void sample(
        self,
        TSArray X,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        double *weights,
        Py_ssize_t length,
        double *mean,
        uint32_t *seed
    ) noexcept nogil:
        cdef Py_ssize_t i
        mean[0] = 0
        for i in range(length):
            weights[i] = rand_normal(self.mean, self.scale, seed)
            mean[0] += weights[i]

        mean[0] = mean[0] / length


cdef class UniformKernelSampler(KernelSampler):

    cdef double lower
    cdef double upper

    def __init__(self, lower=-1.0, upper=1.0):
        self.lower = lower
        self.upper = upper

    cdef void sample(
        self,
        TSArray X,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        double *weights,
        Py_ssize_t length,
        double *mean,
        uint32_t *seed
    ) noexcept nogil:
        cdef Py_ssize_t i
        mean[0] = 0
        for i in range(length):
            weights[i] = rand_uniform(self.lower, self.upper, seed)
            mean[0] += weights[i]

        mean[0] = mean[0] / length


cdef class ShapeletKernelSampler(KernelSampler):
    cdef void sample(
        self,
        TSArray X,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        double *weights,
        Py_ssize_t length,
        double *mean,
        uint32_t *seed
    ) noexcept nogil:
        cdef Py_ssize_t start
        cdef Py_ssize_t index
        cdef Py_ssize_t dim
        cdef Py_ssize_t i

        start = rand_int(0, X.shape[2] - length, seed)
        index = samples[rand_int(0, n_samples, seed)]
        if X.shape[1] > 1:
            dim = rand_int(0, X.shape[1], seed)
        else:
            dim = 1

        mean[0] = 0
        cdef const double *data = &X[index, dim, start]
        for i in range(length):
            weights[i] = data[i]
            mean[0] += weights[i]

        mean[0] /= length


cdef class RocketAttributeGenerator(AttributeGenerator):
    cdef Py_ssize_t n_kernels
    cdef KernelSampler weight_sampler
    cdef double padding_prob
    cdef double bias_prob
    cdef double normalize_prob
    cdef Py_ssize_t *kernel_size
    cdef Py_ssize_t n_kernel_size

    def __cinit__(
        self,
        Py_ssize_t n_kernels,
        KernelSampler weight_sampler,
        object kernel_size,
        double bias_prob,
        double padding_prob,
        double normalize_prob
    ):
        self.n_kernels = n_kernels
        self.weight_sampler = weight_sampler
        self.kernel_size = <Py_ssize_t*> malloc(sizeof(Py_ssize_t) * len(kernel_size))
        self.n_kernel_size = len(kernel_size)

        cdef Py_ssize_t i
        for i in range(len(kernel_size)):
            self.kernel_size[i] = kernel_size[i]

        self.bias_prob = bias_prob
        self.padding_prob = padding_prob
        self.normalize_prob = normalize_prob

    def __reduce__(self):
        return self.__class__, (
            self.n_kernels,
            self.weight_sampler,
            to_ndarray_int(self.kernel_size, self.n_kernel_size),
            self.bias_prob,
            self.padding_prob,
            self.normalize_prob
        )

    def __dealloc__(self):
        free(self.kernel_size)

    cdef Py_ssize_t get_n_attributess(self, TSArray X) noexcept nogil:
        return self.n_kernels

    cdef Py_ssize_t get_n_outputs(self, TSArray X) noexcept nogil:
        return self.get_n_attributess(X) * 2

    cdef Py_ssize_t next_attribute(
        self,
        Py_ssize_t attribute_id,
        TSArray X,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        Attribute *transient,
        uint32_t *seed
    ) noexcept nogil:
        cdef Rocket *rocket = <Rocket*> malloc(sizeof(Rocket))
        cdef Py_ssize_t i
        cdef double mean
        cdef Py_ssize_t length = self.kernel_size[rand_int(0, self.n_kernel_size, seed)]
        cdef double* weight = <double*> malloc(sizeof(double) * length)

        self.weight_sampler.sample(
            X, samples, n_samples, weight, length, &mean, seed
        )
        if rand_uniform(0, 1, seed) < self.normalize_prob:
            for i in range(length):
                weight[i] -= mean

        rocket.length = length
        rocket.dilation = <Py_ssize_t> floor(
            pow(2, rand_uniform(0, log2((X.shape[2] - 1) / (rocket.length - 1)), seed))
        )
        rocket.padding = 0
        if rand_uniform(0, 1, seed) < self.padding_prob:
            rocket.padding = ((rocket.length - 1) * rocket.dilation) // 2

        rocket.return_mean = rand_uniform(0, 1, seed) < 0.5
        rocket.weight = weight

        rocket.bias = 0
        if rand_uniform(0, 1, seed) < self.bias_prob:
            rocket.bias = rand_uniform(-1, 1, seed)

        if X.shape[1] > 1:
            transient.dim = rand_int(0, X.shape[1], seed)
        else:
            transient.dim = 0

        transient.attribute = rocket
        return 0

    cdef Py_ssize_t free_transient(self, Attribute *attribute) noexcept nogil:
        return self.free_persistent(attribute)

    cdef Py_ssize_t free_persistent(self, Attribute *attribute) noexcept nogil:
        cdef Rocket *rocket
        if attribute.attribute != NULL:
            rocket = <Rocket*> attribute.attribute
            if rocket.weight != NULL:
                free(rocket.weight)
            free(attribute.attribute)
            attribute.attribute = NULL
        return 0

    # NOTE: We move ownership of `transient.attribute` to `persistent.attribute`.
    cdef Py_ssize_t init_persistent(
        self,
        TSArray X,
        Attribute *transient,
        Attribute *persistent
    ) noexcept nogil:
        persistent.dim = transient.dim
        persistent.attribute = transient.attribute
        transient.attribute = NULL
        return 0

    cdef double transient_value(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample
    ) noexcept nogil:
        cdef double mean_val, max_val
        cdef Rocket* rocket = <Rocket*> attribute.attribute

        # TODO: (1.3) use utils._cconv.convolution_1d
        apply_convolution(
            rocket.dilation,
            rocket.padding,
            rocket.bias,
            rocket.weight,
            rocket.length,
            &X[sample, attribute.dim, 0],
            X.shape[2],
            &mean_val,
            &max_val,
        )
        if rocket.return_mean:
            return mean_val
        else:
            return max_val

    cdef double persistent_value(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample
    ) noexcept nogil:
        return self.transient_value(attribute, X, sample)

    cdef Py_ssize_t transient_fill(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample,
        double[:, :] out,
        Py_ssize_t out_sample,
        Py_ssize_t attribute_id,
    ) noexcept nogil:
        cdef double mean_val, max_val
        cdef Rocket* rocket = <Rocket*> attribute.attribute
        apply_convolution(
            rocket.dilation,
            rocket.padding,
            rocket.bias,
            rocket.weight,
            rocket.length,
            &X[sample, attribute.dim, 0],
            X.shape[2],
            &mean_val,
            &max_val,
        )
        cdef Py_ssize_t attribute_offset = attribute_id * 2
        out[out_sample, attribute_offset] = mean_val
        out[out_sample, attribute_offset + 1] = max_val
        return 0

    cdef Py_ssize_t persistent_fill(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample,
        double[:, :] out,
        Py_ssize_t out_sample,
        Py_ssize_t out_attribute,
    ) noexcept nogil:
        return self.transient_fill(
            attribute, X, sample, out, out_sample, out_attribute
        )

    cdef object persistent_to_object(self, Attribute *attribute):
        cdef Py_ssize_t j
        cdef Rocket *rocket = <Rocket*> attribute.attribute

        weights = np.empty(rocket.length, dtype=float)
        for j in range(rocket.length):
            weights[j] = rocket.weight[j]

        return attribute.dim, (
            rocket.length,
            rocket.dilation,
            rocket.padding,
            weights,
            rocket.bias,
            rocket.return_mean
        )

    cdef Py_ssize_t persistent_from_object(self, object object, Attribute *attribute):
        dim, (length, dilation, padding, weight, bias, return_mean) = object

        cdef Rocket *rocket = <Rocket*> malloc(sizeof(Rocket))
        rocket.length = length
        rocket.dilation = dilation
        rocket.padding = padding
        rocket.bias = bias
        rocket.weight = <double*> malloc(sizeof(double) * length)
        rocket.return_mean = return_mean

        cdef Py_ssize_t i
        for i in range(length):
            rocket.weight[i] = weight[i]

        attribute.attribute = rocket
        attribute.dim = dim
        return 0

# TODO(1.3): Remove
cdef void apply_convolution(
    Py_ssize_t dilation,
    Py_ssize_t padding,
    double bias,
    double *weight,
    Py_ssize_t length,
    const double* x,
    Py_ssize_t x_length,
    double* mean_val,
    double* max_val
) noexcept nogil:
    cdef Py_ssize_t out_len, end
    cdef Py_ssize_t i, j, k

    out_len = (x_length + 2 * padding) - ((length - 1) * dilation)
    end = (x_length + padding) - ((length - 1) * dilation)
    max_val[0] = -INFINITY
    mean_val[0] = 0.0
    for i in range(-padding, end):
        inner_prod = bias
        k = i
        for j in range(length):
            if -1 < k < x_length:
                inner_prod += weight[j] * x[k]
            k += dilation
        if inner_prod > max_val[0]:
            max_val[0] = inner_prod
        if inner_prod > 0:
            mean_val[0] += 1

    mean_val[0] /= out_len
